#define BLOOM_SIZE   64
int r_bloomtexture = 0;
extern gltexture_t *r_blacktexture;
unsigned int bloomdata[BLOOM_SIZE * BLOOM_SIZE];
unsigned int bloomblur[BLOOM_SIZE * BLOOM_SIZE];
cvar_t r_bloom = {"r_bloom", "2", true};

// fast blur filter - adapted from pseudocode at http://www.blackpawn.com/texts/blur/default.html
int divby[BLOOM_SIZE];

void BlurHorizontal (unsigned int *src, unsigned int *dst, int radius)
{
   int x, y;
   int kx;

   for (y = 0; y < BLOOM_SIZE; y++)
   {
      int total[3] = {0, 0, 0};
      int row = y * BLOOM_SIZE;

      // start at zero because we can't have negative x!!!
      for (kx = 0; kx <= radius; kx++)
      {
         int col = row + kx;

         total[0] += (int) ((unsigned char *) &src[col])[0];
         total[1] += (int) ((unsigned char *) &src[col])[1];
         total[2] += (int) ((unsigned char *) &src[col])[2];
      }

      ((unsigned char *) &dst[row])[0] = total[0] / divby[0];
      ((unsigned char *) &dst[row])[1] = total[1] / divby[0];
      ((unsigned char *) &dst[row])[2] = total[2] / divby[0];

      for (x = 1; x < BLOOM_SIZE; x++)
      {
         // don't bother if too small or too large
         if (x - radius - 1 >= 0)
         {
            int col = row + x - radius - 1;

            total[0] -= (int) ((unsigned char *) &src[col])[0];
            total[1] -= (int) ((unsigned char *) &src[col])[1];
            total[2] -= (int) ((unsigned char *) &src[col])[2];
         }

         if (x + radius < BLOOM_SIZE)
         {
            int col = row + x + radius;

            total[0] += (int) ((unsigned char *) &src[col])[0];
            total[1] += (int) ((unsigned char *) &src[col])[1];
            total[2] += (int) ((unsigned char *) &src[col])[2];
         }

         ((unsigned char *) &dst[row + x])[0] = total[0] / divby[x];
         ((unsigned char *) &dst[row + x])[1] = total[1] / divby[x];
         ((unsigned char *) &dst[row + x])[2] = total[2] / divby[x];
      }
   }
}


void Transpose (unsigned int *src)
{
   int y;

   for (y = 0; y < BLOOM_SIZE; y++)
   {
      int x = y + 1;

      for (; x < BLOOM_SIZE; x++)
      {
         unsigned int tmp = src[y * BLOOM_SIZE + x];
         src[y * BLOOM_SIZE + x] = src[x * BLOOM_SIZE + y];
         src[x * BLOOM_SIZE + y] = tmp;
      }
   }
}


void DoBlur (unsigned int *src, unsigned int *dst, int radius)
{
   int i, j;

   // division factors for handling edge cases smoothly
   // need to recalc each time because of variable radius
   for (i = BLOOM_SIZE - 1, j = 0; ; i--, j++)
   {
      if (j > i) break;

      divby[i] = radius + 1 + j;
      divby[j] = radius + 1 + j;

      if (divby[i] > radius * 2 + 1) divby[i] = radius * 2 + 1;
      if (divby[j] > radius * 2 + 1) divby[j] = radius * 2 + 1;
   }

   // blur horizontally into dst
   BlurHorizontal (src, dst, radius);

   // flip dst
   Transpose (dst);

   // blur horizontally back into src
   BlurHorizontal (dst, src, radius);

   // flip src
   Transpose (src);
}


// projection matrixes for world and bloom overlay
extern float r_world_project[];
float r_bloom_project[16];
qboolean initbloom = false;

byte bloomscaleandbias[256];

void R_InitBloom (void)
{
   // IMPORTANT - we can't call this until after our first call to GL_BeginRendering, as we need to have valid
   // values in glwidth and glheight, so we use a JIT technique instead of calling it during general init.
   // you can just put a call to it in gl_screen.c after calling GL_BeginRendering.
   if (initbloom) return;

   // set up projection matrixes
   // the spec guarantees a stack depth of at least 2
   glMatrixMode (GL_PROJECTION);
   glPushMatrix ();

   // overlay matrix
   glLoadIdentity ();
   glOrtho (0, glwidth, glheight, 0, -99999, 99999);
   glGetFloatv (GL_PROJECTION_MATRIX, r_bloom_project);

   // restore - the world projection matrix is calculated per frame so we don't do it here
   glPopMatrix ();
   glMatrixMode (GL_MODELVIEW);

   // allocate a texture for the bloom picture
   r_bloomtexture = GL_AllocTexture ();

   // bind
   glBindTexture (GL_TEXTURE_2D, r_bloomtexture);

   // upload for the first time with dummy data
   GL_UploadTexture ((byte *) bloomdata, NULL, BLOOM_SIZE, BLOOM_SIZE, false, 32);

   // bloom is initialized now
   initbloom = true;
}


// hack - see below
qboolean firstrun = true;

void R_CaptureBloom (void)
{
   int i;
   int baseradius = 2;
   msurface_t *surf;
   texture_t *tex;

   // no fullbrights to render
   if (!(r_renderflags & RF_FULLBRIGHT)) return;
   if (!r_bloom.value) return;

   // create the viewport to capture to
   R_Set3DViewport (BLOOM_SIZE, BLOOM_SIZE);

   // hack the depthrange to prevent having to clear
   glDepthRange (0.8, 1.0);

   // set up our vertex arrays for the bloom capture
   R_ActivateTMU (GL_TEXTURE0, GL_REPLACE, surfverts[0].st, 0);

   for (tex = texturelist; tex; tex = tex->texturelist)
   {
      // no surfs!
      if (!(surf = tex->texturechain)) continue;

      // not included in bloom
      if (surf->flags & SURF_DRAWSKY) continue;

      // bind the fullbright if it's present, otherwise bind a black texture to ensure that depthmasking works properly.
      // bind black only on first run - filthy evil disgusting hack - see below...
      // important - also do liquids - lava and slime bloom, others don't but must be rendered so that we have depth in the event
      // of having a bsp with opaque water vis...
      if (firstrun)
         glBindTexture (GL_TEXTURE_2D, r_blacktexture->texnum);
      else if ((surf->flags & SURF_DRAWLAVA) || (surf->flags & SURF_DRAWSLIME))
         glBindTexture (GL_TEXTURE_2D, tex->glt->texnum);
      else if (tex->glt->fullbright && !(surf->flags & SURF_DRAWTURB))
         glBindTexture (GL_TEXTURE_2D, tex->glt->fullbright);
      else glBindTexture (GL_TEXTURE_2D, r_blacktexture->texnum);

      // render the texturechain - don't bother warping liquid verts here...
      for (; surf; surf = surf->texturechain) R_RenderSurfacePolygon (surf, RF_TEXST);
   }

   // what about other stuff?  models?  lightmaps?  etc?

   // restore depthrange
   glDepthRange (0.0f, 1.0f);

   // finished rendering
   R_DeactivateTMU (GL_TEXTURE0);

   // capture the bloom texture
   glReadPixels (0, 0, BLOOM_SIZE, BLOOM_SIZE, GL_RGBA, GL_UNSIGNED_BYTE, bloomdata);

   // run the blur on it - this blurs back into bloomdata as a two pass op
   // note - this is done in software for performance reasons!!!  I had a version entirely on the GPU but it was
   // prohibitively slow.  like, it halved frame rates kinda-slow.  so much for "everything on the GPU", eh?
   for (i = 0; i < (int) r_bloom.value; i++)
   {
      DoBlur (bloomdata, bloomblur, baseradius);
      baseradius++;
   }

   // first run hack... the texture is not actually all "black" in the background - there are some patches of 8/4/10 where
   // the so-called "blacktexture" is used (despite being uploaded as black).  this is probably implementation dependent, so
   // we need to read back this texture, find the maximum amount to bias by (for this implementation), and use that.
   if (firstrun)
   {
      int bias = 0;

      for (i = BLOOM_SIZE * BLOOM_SIZE - 1; i >= 0; i--)
      {
         if (((unsigned char *) &bloomdata[i])[0] > bias) bias = ((unsigned char *) &bloomdata[i])[0];
         if (((unsigned char *) &bloomdata[i])[1] > bias) bias = ((unsigned char *) &bloomdata[i])[1];
         if (((unsigned char *) &bloomdata[i])[2] > bias) bias = ((unsigned char *) &bloomdata[i])[2];
      }

      // now we create the scale and bias table, with the derived max to use as a bias factor
      for (i = 0; i < 256; i++)
      {
         // copy to float
         float j = i;

         // bias first
         j -= bias;

         // correct scale range - our blur is mathematically correct, but too faint!!!
         j *= 512.0f;
         j /= (float) (256 - bias);

         // clamp
         if (j < 0) j = 0;
         if (j > 255) j = 255;

         // store back
         bloomscaleandbias[i] = (byte) j;
      }

      // not first run any more
      firstrun = false;
   }

   // pixel transfer ops on the GPU are too expensive, so run scale and bias in software also
   // scale and bias back into bloomblur to avoid head-scratching moments
   for (i = BLOOM_SIZE * BLOOM_SIZE - 1; i >= 0; i--)
   {
      ((unsigned char *) &bloomblur[i])[0] = bloomscaleandbias[((unsigned char *) &bloomdata[i])[0]];
      ((unsigned char *) &bloomblur[i])[1] = bloomscaleandbias[((unsigned char *) &bloomdata[i])[1]];
      ((unsigned char *) &bloomblur[i])[2] = bloomscaleandbias[((unsigned char *) &bloomdata[i])[2]];
   }
}


void R_RenderBloom (void)
{
   // no fullbrights == no blooms
   if (!(r_renderflags & RF_FULLBRIGHT)) return;
   if (!r_bloom.value) return;

   // set up orthographic projection
   glMatrixMode (GL_PROJECTION);
   glLoadMatrixf (r_bloom_project);

   // use identity modelview
   glMatrixMode (GL_MODELVIEW);
    glLoadIdentity ();

   // bind
   glBindTexture (GL_TEXTURE_2D, r_bloomtexture);

   // upload to OpenGL - bloomblur is the final destination
   glTexSubImage2D (GL_TEXTURE_2D, 0, 0, 0, BLOOM_SIZE, BLOOM_SIZE, GL_RGBA, GL_UNSIGNED_BYTE, bloomblur);

   // additive blend
   glEnable (GL_BLEND);
   glBlendFunc (GL_ONE, GL_ONE);
   glDepthMask (GL_FALSE);

   // render
   R_ActivateTMU (GL_TEXTURE0, GL_REPLACE, surfverts[0].st, 0);

   // left/top verts
   surfverts[0].v[0] = surfverts[3].v[0] = 0;
   surfverts[0].v[1] = surfverts[1].v[1] = 0;

   // v[2] is always 0
   surfverts[0].v[2] = surfverts[1].v[2] = surfverts[2].v[2] = surfverts[3].v[2] = 0;

   // right/bottom verts
   surfverts[1].v[0] = surfverts[2].v[0] = glwidth;
   surfverts[2].v[1] = surfverts[3].v[1] = glheight;

   // texcoords
   surfverts[0].st[0] = surfverts[2].st[1] = surfverts[3].st[0] = surfverts[3].st[1] = 0;
   surfverts[0].st[1] = surfverts[1].st[0] = surfverts[1].st[1] = surfverts[2].st[0] = 1;

   glDrawArrays (GL_QUADS, 0, 4);

   R_DeactivateTMU (GL_TEXTURE0);

   // regular blend
   glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
   glDisable (GL_BLEND);
   glDepthMask (GL_TRUE);

   // restore 3d
   glMatrixMode (GL_PROJECTION);
   glLoadMatrixf (r_world_project);

   glMatrixMode (GL_MODELVIEW);
   glLoadMatrixf (r_world_matrix);
}